################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################

import string
from math import sqrt
from typing import Dict, List, Optional


import torch


class Linear(torch.nn.Module):
    def __init__(
        self,
        in_dim: int,
        out_dim: int,
        bias: bool = False,
        l: int = -1,
        **kwargs,
    ) -> None:
        super().__init__()
        assert l >= 0
        self.in_dim = in_dim
        self.out_dim = out_dim
        self.alpha = 1.0 / sqrt(in_dim)
        self.l = l
        self.weight = torch.nn.Parameter(torch.empty((in_dim, out_dim)))
        torch.nn.init.uniform_(self.weight, -sqrt(3), sqrt(3))
        if bias and l == 0:
            self.bias = torch.nn.Parameter(torch.zeros(out_dim))
        else:
            self.register_parameter("bias", None)

        letters = [letter for letter in list(string.ascii_letters)[3:] if letter != 'C']
        in1 = 'bc' + ''.join(letters[:self.l])
        in2 = 'cC'
        out = 'bC' + ''.join(letters[:self.l])
        self.expr = f'{in1}, {in2} -> {out}'
        
    def forward(self, t: torch.Tensor, node_attrs: Optional[torch.Tensor] = None) -> torch.Tensor:
        W = self.weight 
        W = W * self.alpha
        t = torch.einsum(self.expr, t, W)
        if self.bias is not None:
            t = t + self.bias.unsqueeze(0)
        return t
    
    # def forward(
    #         self, 
    #         t: torch.Tensor, 
    #         node_attrs: Optional[torch.Tensor] = None, 
    #         s: Optional[slice] = None,
    #         bias: bool = True,
    #     ) -> torch.Tensor:
    #     W = self.weight if s is None else self.weight[s, :]
    #     W = W * self.alpha
    #     t = torch.einsum(self.expr, t, W)
    #     if self.bias is not None and bias:
    #         t = t + self.bias.unsqueeze(0)
    #     return t

    def __repr__(self):
        return f"{self.__class__.__name__}(in_dim={self.in_dim}, out_dim={self.out_dim}, bias={self.bias is not None}, l={self.l})"


class ElementLinear(torch.nn.Module):
    def __init__(
        self,
        in_dim: int,
        out_dim: int,
        bias: bool = False,
        l: int = -1,
        atomic_numbers: List[int] = None,
        groups: Optional[List[List[int]]] = None,
        **kwargs,
    ):
        super().__init__()
        assert l >= 0
        assert atomic_numbers is not None
        num_elements = len(atomic_numbers)
        self.register_buffer(
            "num_elements", torch.tensor(num_elements, dtype=torch.int64)
        )
        self.in_dim = in_dim
        self.out_dim = out_dim
        self.alpha = 1.0 / sqrt(in_dim)
        self.l = l
        self.atomic_numbers = atomic_numbers
        self.groups = groups
        self.num_groups = len(groups) if groups is not None else None
        # if self.groups is None:
        self.weights = torch.nn.Parameter(
            torch.empty(num_elements, out_dim, in_dim)
        )
        torch.nn.init.uniform_(self.weights, -sqrt(3), sqrt(3))
        if bias and l == 0:
            self.bias = torch.nn.Parameter(torch.empty(num_elements, out_dim))
            torch.nn.init.zeros_(self.bias)
        else:
            self.register_parameter("bias", None)
        
        letters = [c for c in string.ascii_letters[3:] if c not in ['C', 'z']]
        self.expr = (
            f'bz, zCc, bc{"".join(letters[:self.l])} -> '
            f'bC{"".join(letters[:self.l])}'
        )
        # else:
        #     assert sum(len(g) for g in groups) == num_elements
        #     elem2gid = {}
        #     for gid, g in enumerate(groups):
        #         for elem in g:
        #             elem2gid[elem] = gid
        #     gid_map = [elem2gid[z] for z in atomic_numbers]
        #     self.register_buffer("elem2gid", torch.tensor(gid_map, dtype=torch.int64))
        #     self.weights = nn.Parameter(
        #         torch.empty(self.num_groups, out_dim, in_dim)
        #     )
        #     torch.nn.init.uniform_(self.weights, -sqrt(3), sqrt(3))
        #     if bias:
        #         self.bias = nn.Parameter(torch.empty(self.num_groups, out_dim))
        #         torch.nn.init.zeros_(self.bias)
        #     else:
        #         self.register_parameter("bias", None)


    def forward(self, t: torch.Tensor, node_attrs: Optional[torch.Tensor] = None) -> torch.Tensor:
        W = self.weights * self.alpha
        t = torch.einsum(self.expr, node_attrs, W, t)
        if self.bias is not None:
            b = torch.einsum('bz, zC -> bC', node_attrs, self.bias)
            t = t + b
        return t

    def __repr__(self):
        return (f"{self.__class__.__name__}(in_dim={self.in_dim}, "
                    f"out_dim={self.out_dim}, bias={self.bias is not None}, l={self.l})")


# only for prod
class CWLinear(torch.nn.Module):
    def __init__(
        self,
        in_dim: int,
        out_dim: int,
        bias: bool = False,
        l: int = -1,
        **kwargs,
    ):
        super().__init__()
        assert l >= 0
        self.l = l
        self.in_dim = in_dim
        self.out_dim = out_dim
        self.channel = out_dim
        num_path = int(in_dim / out_dim)
        self.num_path = num_path
        self.alpha = 1.0 / sqrt(num_path)
        self.weight = torch.nn.Parameter(torch.empty(out_dim, num_path))
        torch.nn.init.uniform_(self.weight, -sqrt(3), sqrt(3))
        if bias and self.l == 0:
            self.bias = torch.nn.Parameter(torch.zeros(out_dim))
        else:
            self.register_parameter("bias", None)
        letters = list(string.ascii_letters[3:])
        self.expr = (
            f'abc{"".join(letters[:self.l])}, ca -> '
            f'bc{"".join(letters[:self.l])}'
        )

    def forward(self, t: torch.Tensor, node_attrs: Optional[torch.Tensor] = None) -> torch.Tensor:
        W = self.weight * self.alpha
        t = torch.einsum(self.expr, t, W)
        if self.bias is not None:
            t = t + self.bias.unsqueeze(0)
        return t

    def __repr__(self):
        return f"{self.__class__.__name__}(in_dim={self.num_path}, out_dim={self.channel}, bias={self.bias is not None}, l={self.l})"


# only for prod
class ElementCWLinear(torch.nn.Module):
    def __init__(
        self,
        in_dim: int,
        out_dim: int,
        bias: bool = False,
        l: int = -1,
        atomic_numbers: List[int] = None,
        **kwargs,
    ):
        super().__init__()
        assert l >= 0
        assert atomic_numbers is not None
        num_elements = len(atomic_numbers)
        self.register_buffer(
            "num_elements", torch.tensor(num_elements, dtype=torch.int64)
        )
        self.in_dim = in_dim
        self.out_dim = out_dim
        self.channel = out_dim
        num_path = int(in_dim / out_dim)
        self.num_path = num_path
        self.atomic_numbers = atomic_numbers
        self.alpha = 1.0 / sqrt(num_path)
        self.l = l
        self.weights = torch.nn.Parameter(torch.empty(num_elements, out_dim, num_path))
        torch.nn.init.uniform_(self.weights, -sqrt(3), sqrt(3))
        if bias and l == 0:
            self.bias = torch.nn.Parameter(torch.zeros(num_elements, out_dim))
        else:
            self.register_parameter("bias", None)
        letters = [c for c in string.ascii_letters[3:] if c != 'z']
        self.expr = (
            f'bz, zca, abc{"".join(letters[:self.l])} -> '
            f'bc{"".join(letters[:self.l])}'
        )

    def forward(self, t: torch.Tensor, node_attrs: Optional[torch.Tensor] = None) -> torch.Tensor:
        W = self.weights * self.alpha
        t = torch.einsum(self.expr, node_attrs, W, t)
        if self.bias is not None:
            b = torch.einsum('bz, zC -> bC')
            t = t + b
        return t

    def __repr__(self):
        return (f"{self.__class__.__name__}(in_dim={self.num_path}, "
                    f"out_dim=1, bias={self.bias is not None}, l={self.l})")


class SelfInteraction(torch.nn.Module):

    ls: List[int]

    def __init__(
        self,
        in_channel: int | List[int],
        out_channel: int | List[int],
        ls: List[int],
        bias: bool = False,
        atomic_numbers: Optional[List[int]] = None,
    ) -> None:
        super().__init__()

        self.ls = ls

        if isinstance(in_channel, int):
            in_channel = [in_channel] *len(ls)

        if isinstance(out_channel, int):
            out_channel = [out_channel] *len(ls)

        if atomic_numbers is None:
            self.linears = torch.nn.ModuleDict(
                {
                    str(l): Linear(
                        in_channel[idx],
                        out_channel[idx],
                        bias=(l == 0 and bias),
                        l=l,
                    )
                    for idx, l in enumerate(ls)
                }
            )
        else:
            self.linears = torch.nn.ModuleDict(
                {
                    str(l): ElementLinear(
                        in_channel[idx],
                        out_channel[idx],
                        bias=(l == 0 and bias),
                        l=l,
                        atomic_numbers=atomic_numbers,
                    )
                    for idx, l in enumerate(ls)
                }
            )

    def forward(
            self, 
            in_dict: Dict[int, torch.Tensor],
            node_attrs: Optional[torch.Tensor] = None,
        ) -> Dict[int, torch.Tensor]:
        out_dict = {}
        for l in self.ls:
            out_dict[l] = self.linears[str(l)](in_dict[l], node_attrs)
        return out_dict


class LoRALinear(torch.nn.Module):
    def __init__(
        self,
        in_dim: int,
        out_dim: int,
        bias: bool = False,
        l: int = -1,
        lora_r: int = 8,
        lora_alpha: float = 8.0,
        **kwargs,
    ):
        super().__init__()
        assert l >= 0
        self.in_dim = in_dim
        self.out_dim = out_dim
        self.l = l
        self.alpha = 1.0 / sqrt(in_dim)
        self.use_lora = True
        self.lora_r = lora_r
        self.lora_alpha = lora_alpha
        self.scaling = lora_alpha / lora_r
        self.weight = torch.nn.Parameter(torch.empty(in_dim, out_dim))
        torch.nn.init.uniform_(self.weight, -sqrt(3), sqrt(3))
        self.weight.requires_grad_(False)
        if bias and l == 0:
            self.bias = torch.nn.Parameter(torch.zeros(out_dim))
            self.bias.requires_grad_(False)
        else:
            self.register_parameter("bias", None)
        self.lora_A = torch.nn.Parameter(torch.zeros(in_dim, lora_r))
        self.lora_B = torch.nn.Parameter(torch.zeros(lora_r, out_dim))
        torch.nn.init.kaiming_uniform_(self.lora_A, a=sqrt(5))
        torch.nn.init.zeros_(self.lora_B)

        letters = [c for c in string.ascii_letters[3:] if c != 'C']
        in1 = 'bc' + ''.join(letters[:self.l])
        in2 = 'cC'
        out = 'bC' + ''.join(letters[:self.l])
        self.expr = f'{in1}, {in2} -> {out}'

    def forward(self, t: torch.Tensor, node_attrs: Optional[torch.Tensor] = None) -> torch.Tensor:

        if self.use_lora:
            delta_w = self.lora_A @ (self.lora_B * self.scaling)
            W = self.weight + delta_w
        else:
            W = self.weight

        W = W * self.alpha
        t = torch.einsum(self.expr, t, W)

        if self.bias is not None:
            t = t + self.bias.unsqueeze(0)

        return t
    
    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(\n"
            f"  in_dim      = {self.in_dim},\n"
            f"  out_dim     = {self.out_dim},\n"
            f"  bias        = {self.bias is not None},\n"
            f"  alpha       = {self.alpha:.2f},\n"
            f"  lora_r      = {self.lora_r},\n"
            f"  lora_alpha  = {self.lora_alpha},\n"
            f")"
        )


class LoRAElementLinear(torch.nn.Module):
    def __init__(
        self,
        in_dim: int,
        out_dim: int,
        bias: bool = False,
        l: int = -1,
        atomic_numbers: List[int] = None,
        lora_r: int = 8,
        lora_alpha: float = 8.0,
        element_aware: bool = True, 
        **kwargs,
    ):
        super().__init__()
        assert l >= 0
        assert atomic_numbers is not None
        num_elements = len(atomic_numbers)
        self.register_buffer("num_elements", torch.tensor(num_elements))
        self.in_dim = in_dim
        self.out_dim = out_dim
        self.l = l
        self.alpha = 1.0 / sqrt(in_dim)
        self.atomic_numbers = atomic_numbers
        self.use_lora = True
        self.element_aware = element_aware
        self.lora_r = lora_r
        self.lora_alpha = lora_alpha
        self.scaling = lora_alpha / lora_r
        self.weights = torch.nn.Parameter(
            torch.empty(num_elements, out_dim, in_dim)
        )
        torch.nn.init.uniform_(self.weights, -sqrt(3), sqrt(3))
        self.weights.requires_grad_(False)
        if bias and l == 0:
            self.bias = torch.nn.Parameter(torch.zeros(num_elements, out_dim))
            self.bias.requires_grad_(False)
        else:
            self.register_parameter("bias", None)
        if self.element_aware:
            self.lora_A = torch.nn.Parameter(torch.zeros(num_elements, lora_r, in_dim))
            self.lora_B = torch.nn.Parameter(torch.zeros(num_elements, out_dim, lora_r))
        else:
            self.lora_A = torch.nn.Parameter(torch.zeros(lora_r, in_dim))
            self.lora_B = torch.nn.Parameter(torch.zeros(out_dim, lora_r))
        torch.nn.init.kaiming_uniform_(self.lora_A, a=sqrt(5))
        torch.nn.init.zeros_(self.lora_B)
        letters = [c for c in string.ascii_letters[3:] if c not in ['C', 'z']]
        self.expr = (
            f'bz, zCc, bc{"".join(letters[:self.l])} -> '
            f'bC{"".join(letters[:self.l])}'
        )

    def forward(self, t: torch.Tensor, node_attrs: Optional[torch.Tensor] = None) -> torch.Tensor:
        if self.use_lora:
            if self.element_aware:
                delta_w = torch.einsum(
                    'zri, zor -> zoi',
                    self.lora_A,
                    self.lora_B * self.scaling,
                )
            else:
                delta_w = (self.lora_B * self.scaling) @ self.lora_A 
                delta_w = delta_w.unsqueeze(0)
            W = self.weights + delta_w
        else:
            W = self.weights
        W = W * self.alpha
        t = torch.einsum(self.expr, node_attrs, W, t)
        if self.bias is not None:
            b = torch.einsum('bz, zC -> bC', node_attrs, self.bias)
            t = t + b
        return t

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(\n"
            f"  in_dim             = {self.in_dim},\n"
            f"  out_dim            = {self.out_dim},\n"
            f"  bias               = {self.bias is not None},\n"
            f"  alpha              = {self.alpha:.2f},\n"
            f"  lora_r             = {self.lora_r},\n"
            f"  lora_alpha         = {self.lora_alpha},\n"
            f"  lora_element_aware = {self.element_aware},\n"
            f")"
        )
    
    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(in_dim={self.in_dim}, out_dim={self.out_dim} bias={self.bias is not None}, alpha={self.alpha:.2f}, lora_r={self.lora_r}, lora_alpha={self.lora_alpha}, ) "


# only for prod
class LoRACWLinear(torch.nn.Module):
    def __init__(
        self,
        in_dim: int,
        out_dim: int,
        bias: bool = False,
        l: int = -1,
        lora_r: int = 8,
        lora_alpha: float = 8.0,
        **kwargs,
    ):
        super().__init__()
        assert l >= 0
        self.in_dim = in_dim
        self.out_dim = out_dim
        self.channel = out_dim
        self.num_path = int(in_dim / out_dim)
        self.l = l
        self.alpha = 1.0 / sqrt(self.num_path)
        self.use_lora = True
        self.lora_alpha = lora_alpha
        self.lora_r = lora_r
        self.scaling = lora_alpha / lora_r
        self.weight = torch.nn.Parameter(torch.empty(out_dim, self.num_path))
        torch.nn.init.uniform_(self.weight, -sqrt(3), sqrt(3))
        self.weight.requires_grad_(False)
        if bias and l == 0:
            self.bias = torch.nn.Parameter(torch.zeros(out_dim))
            self.bias.requires_grad_(False)
        else:
            self.register_parameter("bias", None)
        self.lora_A = torch.nn.Parameter(torch.zeros(self.num_path, lora_r))
        self.lora_B = torch.nn.Parameter(torch.zeros(lora_r, out_dim))
        torch.nn.init.kaiming_uniform_(self.lora_A, a=sqrt(5))
        torch.nn.init.zeros_(self.lora_B)
        letters = list(string.ascii_letters[3:])
        self.expr = (
            f'abc{"".join(letters[:self.l])}, ca -> '
            f'bc{"".join(letters[:self.l])}'
        )

    def forward(self, t: torch.Tensor, node_attrs: Optional[torch.Tensor] = None) -> torch.Tensor:
        if self.use_lora:
            delta_w = (self.lora_A @ self.lora_B.T) * self.scaling
            W = self.weight + delta_w
        else:
            W = self.weight
        W = W * self.alpha
        t = torch.einsum(self.expr, t, W)
        if self.bias is not None:
            t = t + self.bias.unsqueeze(0)
        return t

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(\n"
            f"  in_dim      = {self.in_dim},\n"
            f"  out_dim     = {self.out_dim},\n"
            f"  bias        = {self.bias is not None},\n"
            f"  alpha       = {self.alpha:.2f},\n"
            f"  lora_r      = {self.lora_r},\n"
            f"  lora_alpha  = {self.lora_alpha},\n"
            f")"
        )


# only for prod
class LoRAElementCWLinear(torch.nn.Module):
    def __init__(
        self,
        in_dim: int,
        out_dim: int,
        bias: bool = False,
        l: int = -1,
        atomic_numbers: List[int] = None,
        lora_r: int = 8,
        lora_alpha: float = 8.0,
        **kwargs,
    ):
        super().__init__()
        assert l >= 0
        assert atomic_numbers is not None
        self.in_dim = in_dim
        self.out_dim = out_dim
        num_elements = len(atomic_numbers)
        self.register_buffer("num_elements", torch.tensor(num_elements))
        self.channel = out_dim
        self.num_path = int(in_dim / out_dim)
        self.l = l
        self.alpha = 1.0 / sqrt(self.num_path)
        self.atomic_numbers = atomic_numbers
        self.use_lora = True
        self.lora_r = lora_r
        self.lora_alpha = lora_alpha
        self.scaling = lora_alpha / lora_r
        self.weights = torch.nn.Parameter(
            torch.empty(num_elements, out_dim, self.num_path)
        )
        torch.nn.init.uniform_(self.weights, -sqrt(3), sqrt(3))
        self.weights.requires_grad_(False)
        if bias and l == 0:
            self.bias = torch.nn.Parameter(torch.zeros(num_elements, out_dim))
            self.bias.requires_grad_(False)
        else:
            self.register_parameter("bias", None)

        self.lora_A = torch.nn.Parameter(torch.zeros(self.num_path, lora_r))
        self.lora_B = torch.nn.Parameter(torch.zeros(lora_r, out_dim))
        torch.nn.init.kaiming_uniform_(self.lora_A, a=sqrt(5))
        torch.nn.init.zeros_(self.lora_B)
        letters = [c for c in string.ascii_letters[3:] if c != 'z']
        self.expr = (
            f'bz, zca, abc{"".join(letters[:self.l])} -> '
            f'bc{"".join(letters[:self.l])}'
        )

    def forward(self, t: torch.Tensor, node_attrs: Optional[torch.Tensor] = None) -> torch.Tensor:
        if self.use_lora:
            delta_w = (self.lora_B.T @ self.lora_A.T) * self.scaling
            W = self.weights + delta_w.unsqueeze(0)
        else:
            W = self.weights
        W = W * self.alpha
        t = torch.einsum(self.expr, node_attrs, W, t)
        if self.bias is not None:
            b = torch.einsum('bz, zC -> bC', node_attrs, self.bias)
            t = t + b
        return t

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(\n"
            f"  in_dim      = {self.in_dim},\n"
            f"  out_dim     = {self.out_dim},\n"
            f"  bias        = {self.bias is not None},\n"
            f"  alpha       = {self.alpha:.2f},\n"
            f"  lora_r      = {self.lora_r},\n"
            f"  lora_alpha  = {self.lora_alpha},\n"
            f")"
        )
    

# (elemenet-aware, channel-wise)
LinearDict = {
    (False, False): Linear,
    (True, False): ElementLinear,
    (False, True): CWLinear,
    (True, True): ElementCWLinear,
}

LoRALinearDict = {
    (False, False): LoRALinear,
    (True, False): LoRAElementLinear,
    (False, True): LoRACWLinear,
    (True, True): LoRAElementCWLinear,
}


################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################

import numpy as np
import torch
from torch import Tensor


def default_value_for_rank0_atom(num_atoms: int, class_: str,  **kwargs):
    return np.zeros((num_atoms,))


def default_value_for_rank1_atom(num_atoms: int, class_: str,  **kwargs):
    return np.zeros((num_atoms, 3))


def default_value_for_rank2_atom(num_atoms: int, class_: str,  **kwargs):
    return np.zeros((num_atoms, 3, 3))

def default_value_for_rank3_atom(num_atoms: int, class_: str,  **kwargs):
    return np.zeros((num_atoms, 3, 3, 3))

def default_value_for_rank4_atom(num_atoms: int, class_: str,  **kwargs):
    return np.zeros((num_atoms, 3, 3, 3, 3))

def default_value_for_rank0_graph(num_atoms: int, class_: str, **kwargs):
    return np.zeros((1,))


def default_value_for_rank1_graph(num_atoms: int,  class_: str, **kwargs):
    return np.zeros((3,))


def default_value_for_rank2_graph(num_atoms: int,  class_: str, **kwargs):
    return np.zeros((3, 3))


def default_value_for_rank3_graph(num_atoms: int,  class_: str, **kwargs):
    return np.zeros((3, 3, 3))


def default_value_for_rank4_graph(num_atoms: int,  class_: str, **kwargs):
    return np.zeros((3, 3, 3, 3))


def default_value_for_hessians(num_atoms: int, class_: str, **kwargs):
    return np.zeros((num_atoms * 3 * num_atoms * 3))


def shape_fn_for_hessians(t: Tensor, num_atoms: int, **kwargs):
    # assert np.allclose(t, t.T, atol=1e-6)
    shape1 = (num_atoms * 3 , num_atoms * 3)
    shape2 = (num_atoms * 3 * num_atoms * 3)
    assert t.shape == shape1 or t.shape == shape2, (
        f"hessians shape mismatch: got {tuple(t.shape)}, expected {shape1} or {shape2}"
    )
    return t.view(-1)



def voigt_to_matrix(t: Tensor, **kwargs):
    """
    Convert voigt notation to matrix notation
    """
    if t.shape == (3, 3):
        return t
    if t.shape == (6,):
        return torch.tensor(
            [
                [t[0], t[5], t[4]],
                [t[5], t[1], t[3]],
                [t[4], t[3], t[2]],
            ],
            dtype=t.dtype,
        )
    if t.shape == (9,):
        return t.view(3, 3)

    raise ValueError(
        f"Stress tensor must be of shape (6,) or (3, 3), or (9,) but has shape {t.shape}"
    )



################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################

from typing import Optional, Tuple

from ase.geometry import complete_cell

import numpy as np
from matscipy.neighbours import neighbour_list
try:
    from ase.neighborlist import primitive_neighbor_list
    from vesin import NeighborList as vesin_nl
except ImportError:
    pass


# self-interaction: ase
# 1D, 2D: ase, matscipy
NL_BACKEND = ["ase", "vesin", "matscipy"]


def filter_max_neighbors(source, target, shifts, distances, max_neighbors="inf"):

    if max_neighbors is None or max_neighbors == "inf":
        return source, target, shifts
    order = np.lexsort((distances, source))
    src_sorted = source[order]
    dst_sorted = target[order]
    shifts_sorted = shifts[order]

    unique_src, counts = np.unique(src_sorted, return_counts=True)
    cum_counts = np.cumsum(counts)  # [3, 2, 1] => [3, 5, 6]

    mask = np.zeros(len(src_sorted), dtype=bool)
    start_idx = 0
    for end_idx in cum_counts:
        count = end_idx - start_idx
        keep = min(max_neighbors, count)
        mask[start_idx : start_idx + keep] = True
        start_idx = end_idx

    return (
        src_sorted[mask],
        dst_sorted[mask],
        shifts_sorted[mask],
    )

def get_neighborhood(
    positions: np.ndarray,
    cutoff: float,
    pbc: Optional[bool | Tuple[bool, bool, bool]] = None,
    lattice: Optional[np.ndarray] = None,  # [3, 3]
    max_neighbors: Optional[int] = None,
    backend: str = "vesin" # "matscipy",
) -> Tuple[np.ndarray, np.ndarray]:
    
    assert backend in NL_BACKEND, f"Neighborlist backend should be in {NL_BACKEND}"

    # === PBC ===
    if pbc is None:
        pbc = (False, False, False)
    elif isinstance(pbc, bool):
        pbc = (pbc,) * 3
    else:
        pbc = tuple(bool(i) for i in pbc)

    # === Lattiace ===
    if any(pbc):
        if lattice is None or np.allclose(lattice, 0.0):
            raise ValueError(
                "At least one direction is periodic, but lattice is None or zero."
            )
    if not any(pbc):
        lattice = None
    else:
        if lattice is None:
            raise ValueError(
                "At least one direction is periodic, but lattice is None."
            )
        
    # === Neighborlist ===
    if backend == "matscipy":
        edges = neighbour_list(
            quantities="ijSd",
            pbc=pbc,
            cell=lattice,
            positions=positions,
            cutoff=cutoff,
        )
    elif backend == "vesin":
        # https://github.com/Luthaf/vesin/blob/main/python/vesin/src/vesin/_ase.py
        if all(pbc):
            is_3D = True
        elif not any(pbc):
            is_3D = False
        else:
            raise ValueError("vesin only support pbc=(F, F, F) or (T, T, T)")
        if lattice is None:
            box = np.zeros((3, 3), dtype=positions.dtype)
        else:
            box = lattice
        edges = vesin_nl(
            cutoff=cutoff, 
            full_list=True
        ).compute(points=positions, box=box, periodic=is_3D, quantities="ijSd")
        edges = list(edges)
        edges[0] = edges[0].astype(np.int64)
        edges[1] = edges[1].astype(np.int64)
        edges = tuple(edges)
    elif backend == "ase":
        if lattice is None:
            cell = np.zeros((3, 3), dtype=positions.dtype)
        else:
            cell = lattice
        edges = primitive_neighbor_list(
            "ijSd",
            pbc,
            cell,
            positions,
            cutoff=cutoff,
            self_interaction=False,
            use_scaled_positions=False,
        )

    source, target, shifts = filter_max_neighbors(*edges, max_neighbors=max_neighbors)

    real_self_loop = source == target
    real_self_loop &= np.all(shifts == 0, axis=1)
    keep_edge = ~real_self_loop

    source = source[keep_edge]
    target = target[keep_edge]

    edge_shifts = shifts[keep_edge]
    edge_index = np.stack((source, target))

    if lattice is None:
        lattice = np.zeros((3, 3), dtype=positions.dtype)
 
    return edge_index, edge_shifts, pbc, lattice


def get_neighborhood_for_calculator(
    positions: np.ndarray,
    cutoff: float,
    pbc: Optional[Tuple[bool, bool, bool]] = None,
    lattice: Optional[np.ndarray] = None,  # [3, 3]
    max_neighbors: Optional[int] = None,
) -> Tuple[np.ndarray, np.ndarray]:

    edges = neighbour_list(
        quantities="ijSd",
        pbc=pbc,
        cell=lattice,
        positions=positions,
        cutoff=cutoff,
    )
    source, target, shifts = filter_max_neighbors(*edges, max_neighbors=max_neighbors)

    real_self_loop = source == target
    real_self_loop &= np.all(shifts == 0, axis=1)
    keep_edge = ~real_self_loop

    source = source[keep_edge]
    target = target[keep_edge]

    edge_shifts = shifts[keep_edge]
    edge_index = np.stack((source, target))

    return edge_index, edge_shifts






################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################
"""
So far, all the functions have been concentrated within one interaction module,
and LAMMPS utils are modified from MACE
"""
import abc
from typing import Dict, List, Tuple, Optional, Any


import torch
from torch import Tensor


from .mlp import MLP
from .linear import SelfInteraction, LinearDict
from .ctr import Contraction
from .utils import Graph, LAMMPS_MP, dict2flatten, flatten2dict, add_dict_to_left
from .layers import NormNonlinearity, GatedNonlinearity
from .ictd import ICTD
from .utils import expand_dims_to
from ...utils.torch_scatter import scatter_sum


class InteractionBase(torch.nn.Module):
    def __init__(
        self,
        layer: int,
        num_layers: int,
        num_channel: int,
        num_channel_hidden: int,
        bias: bool,
        lmax_in: int,
        lmax_out: int,
        ls_sc: List[int],
        atomic_numbers: int,
        avg_num_neighbors: float,
        num_radial_basis: int,
        radial_mlp: Dict,
        inter: Dict,
    ) -> None:
        super().__init__()
        self.layer = layer
        self.num_layers = num_layers
        self.num_channel = num_channel
        self.num_channel_hidden = num_channel_hidden
        self.bias = bias
        self.lmax_in = lmax_in
        self.lmax_out = lmax_out
        self.ls_sc = ls_sc
        self.atomic_numbers = atomic_numbers
        self.num_radial_basis = num_radial_basis
        self.radial_mlp = radial_mlp
        self.inter = inter
        self.register_buffer(
            "avg_num_neighbors",
            torch.tensor(avg_num_neighbors, dtype=torch.get_default_dtype()),
        )
        for l in range(lmax_out + 1):
            DS = ICTD(l, l)[1]
            self.register_buffer(f"D_{l}_{l}_1", DS[0].to(torch.get_default_dtype()))
            del DS
        self._setup()

    def D(self, l: int):
        return dict(self.named_buffers())[f"D_{l}_{l}_1"]
    
    @abc.abstractmethod
    def _setup(self) -> None:
        raise NotImplementedError

    def handle_lammps(
        self,
        node_feats: Dict[int, Tensor],
        lmp_data: Optional[Any],
        lmp_natoms: Tuple[int, int],
        layer: int,
    ) -> Tensor:  
        _, nghosts = lmp_natoms
        first_layer = (layer == 0)
        if lmp_data is None or first_layer or torch.jit.is_scripting():
            return node_feats
        max_r = max(node_feats.keys())
        node_feats = dict2flatten(max_r, node_feats)
        pad = torch.zeros(
            (nghosts, node_feats.shape[1]),
            dtype=node_feats.dtype,
            device=node_feats.device,
        )
        node_feats = torch.cat((node_feats, pad), dim=0)
        node_feats = LAMMPS_MP.apply(node_feats, lmp_data)
        return flatten2dict(max_r, node_feats, self.num_channel)
    
    def truncate_ghosts(
        self, t: torch.Tensor, nlocal: Optional[int] = None
    ) -> torch.Tensor:
        return t[:nlocal] if nlocal is not None else t
    
    def truncate_ghosts_dict(
        self, t: Dict[int, torch.Tensor], nlocal: Optional[int] = None
    ) -> Dict[int, torch.Tensor]:
        lmax = max(t.keys())
        t = dict2flatten(lmax, t)
        t = self.truncate_ghosts(t, nlocal)
        t = flatten2dict(lmax, t, self.num_channel)
        return t
    
class Interaction(InteractionBase):
    def _setup(self) -> None:
        # === resnet === 
        if self.inter.get('use_resnet', False):
            self.resnet = SelfInteraction(
                self.num_channel,
                self.num_channel,
                ls=list(range(self.lmax_in + 1)),
                bias=self.bias,
                atomic_numbers=self.atomic_numbers,
            )    

        # === linear up ===  
        self.linear_up = SelfInteraction(
            self.num_channel,
            self.num_channel,
            ls=list(range(self.lmax_in + 1)),
            bias=self.bias,
        )

        # === ICTP and ICTC
        self.tc = Contraction(
            num_channel=self.num_channel,
            num_channel_hidden=self.num_channel_hidden,
            lmax_in=self.lmax_in,
            lmax_out=self.lmax_out,
            l1l2=self.inter["l1l2"][self.layer],
            filter_combs=self.inter.get("filter_combs", None)
        )

        # === self / skip connection ===
        self.sc_from = None
        if self.inter.get('sc', {}).get('use_first_sc', False) \
        or self.layer > 0 \
        or self.num_layers == 1:
            self.sc_from = self.inter.get('sc', {}).get('from', "current_message") 
            self.scs = torch.nn.ModuleDict()
            for l in self.ls_sc:
                    self.scs[str(l)] = LinearDict[True, False](
                    self.num_channel if self.sc_from == 'last_product' else self.num_channel_hidden,
                    self.num_channel,
                    bias=(l == 0 and self.bias),
                    atomic_numbers=self.atomic_numbers,
                    l=l,
                )

        # ==== conv weights ====
        conv_weights_type = self.inter.get('conv_weights', ['edge_ij'])
        assert 'edge_ij' in conv_weights_type, 'edge_ij must be in conv_weights' 
        radial_in_dim = self.num_radial_basis + self.num_channel * (len(conv_weights_type)-1)
        need_layer_norm = radial_in_dim != self.num_radial_basis
        self.radial_net = MLP(
            radial_in_dim,
            self.tc.weight_numel,
            self.radial_mlp["hidden"][self.layer],
            act=self.radial_mlp["act"],
            bias=self.radial_mlp.get('bias', False),
            forward_weight_init=True,
            enable_layer_norm=need_layer_norm,
        )
        if 'node_j' in conv_weights_type:
            self.source_embedding = MLP(
                len(self.atomic_numbers),
                self.num_channel,
                hidden_dim=[],
                act=None,
                bias=False,
                forward_weight_init=True,
                enable_layer_norm=False,
            )
            torch.nn.init.uniform_(self.source_embedding.mlp[0].weight, a=-0.001, b=0.001)
        if 'node_i' in conv_weights_type:
            self.target_embedding = MLP(
                len(self.atomic_numbers),
                self.num_channel,
                hidden_dim=[],
                act=None,
                bias=False,
                forward_weight_init=True,
                enable_layer_norm=False,
            )
            torch.nn.init.uniform_(self.target_embedding.mlp[0].weight, a=-0.001, b=0.001)

        # ==== normalizer ====
        self.normalizer = self.inter.get('normalizer', 'avg_num_neighbors')
        assert self.normalizer in ["avg_num_neighbors", "density_v1"]
        if self.normalizer == 'density_v1': # density_v1 are from mace
            self.density_normalizer = MLP(
                radial_in_dim,
                1,
                [64],
                act='silu',
                bias=True,
                forward_weight_init=True,
                enable_layer_norm=need_layer_norm,
            )
            self.alpha = torch.nn.Parameter(torch.tensor(20.0), requires_grad=True)
            self.beta = torch.nn.Parameter(torch.tensor(0.0), requires_grad=True)

        # === nonlinearity ===
        nonlinearity_type = self.inter.get('nonlinearity', {}).get('type', None)
        nonlinearity_gate = self.inter.get('nonlinearity', {}).get('gate', 'silu')
        if nonlinearity_type is not None:
            if nonlinearity_type in ['gated', 'gated_gate']:
                self.nonlinearity_gate = GatedNonlinearity(
                    rmax=self.lmax_out,
                    in_dim=self.num_channel_hidden,
                    gate=nonlinearity_gate,
                )  
            elif nonlinearity_type in ['norm', 'norm_gate']:
                self.nonlinearity_gate = NormNonlinearity(
                    rmax=self.lmax_out,
                    in_dim=self.num_channel_hidden,
                    gate=nonlinearity_gate,
                )
            else:
                raise
            self.nonlinearity_linear = SelfInteraction(
                in_channel=self.num_channel_hidden,
                out_channel=self.num_channel_hidden,
                ls=list(range(self.lmax_out + 1)),
                bias=self.bias,
            )
    
    def forward(
        self,
        node_feats: Dict[int, Tensor],
        node_attrs_total: Tensor,
        node_attrs_slice: Tensor,
        edge_feats: Tensor,
        edge_attrs: Dict[int, Tensor],
        edge_index: Tensor,
        cutoff: Tensor,
        graph: Graph,
    ) -> Tuple[Dict[int, Tensor], Dict[int, Tensor]]:

        # === LAMMPS pretreatment ===
        lmp_data = graph.lmp_data
        lmp_natoms = graph.lmp_natoms
        nlocal = lmp_natoms[0] if lmp_data is not None else None

        scs = {}
        # === self connection / skip connection ===
        if self.sc_from == 'last_product':
            for r in self.ls_sc:
                scs[r] = self.scs[str(r)](m_i[r], node_attrs_total)

        # === residual === 
        residual = {}
        if hasattr(self, 'resnet'):
            residual = self.resnet(node_feats, node_attrs_slice)
            residual = self.handle_lammps(
                residual,
                lmp_data=lmp_data,
                lmp_natoms=lmp_natoms,
                layer=self.layer,
            )

        # === linear up === 
        node_feats = self.linear_up(node_feats)
        node_feats = self.handle_lammps( 
            node_feats,
            lmp_data=lmp_data,
            lmp_natoms=lmp_natoms,
            layer=self.layer,
        )

        # === conv_weights === 
        edge_embedding: list[torch.Tensor] = [edge_feats]
        if hasattr(self, 'source_embedding'):
            edge_embedding.append(self.source_embedding(node_attrs_total)[edge_index[0]])
        if hasattr(self, 'target_embedding'):
            edge_embedding.append(self.target_embedding(node_attrs_total)[edge_index[1]])
        full_edge_feats = torch.cat(edge_embedding, dim=-1)
        conv_weights = self.radial_net(full_edge_feats)
        if cutoff is not None:
            conv_weights = conv_weights * cutoff

        # === normalizer ===
        if hasattr(self, 'density_normalizer'):
            edge_density = torch.tanh(self.density_normalizer(full_edge_feats) ** 2)
            if cutoff is not None:
                edge_density = edge_density * cutoff
            density = scatter_sum(edge_density, edge_index[1], dim=0, dim_size=node_attrs_total.size(0)) 
            density = density * self.beta + self.alpha
            density = density.masked_fill(density == 0, 1e-9)

        # === ICTP and ICTC ===
        r_m_i: Dict[int, torch.Tensor] = self.tc(node_feats, edge_attrs, conv_weights, edge_index)
        m_i = {}
        for r, t in r_m_i.items():
            B = t.size(0)
            C = t.size(1)
            REST = (3,) * r
            if self.normalizer == 'density_v1':
                normalizer = expand_dims_to(density, r+2, dim=-1)
            else:
                normalizer = self.avg_num_neighbors
            t = t / normalizer
            m_i[r] = (t.reshape(B, C, -1) @ self.D(r)).reshape((B, C) + REST)
        add_dict_to_left(m_i, residual)

        # === nonlinearity ===
        if hasattr(self, 'nonlinearity_gate'):
            m_i = self.nonlinearity_linear(self.nonlinearity_gate(m_i))

        # === self connection / skip connection ===
        if self.sc_from == 'current_message':
            for r in self.ls_sc:
                scs[r] = self.scs[str(r)](m_i[r], node_attrs_total)

        # === LAMMPS postprocessing ===
        if graph.lmp:
            m_i = self.truncate_ghosts_dict(m_i, nlocal)
            if len(scs) > 0:
                scs = self.truncate_ghosts_dict(scs, nlocal)
 
        return m_i, scs



